from common import *


def test_corner_1():
    """
    adding isolated vertices will not harm the performance of the algorithm so we can ignore this case
    """
    vertex_num = 4
    weighted_edges = [
        mwpf.HyperEdge([0], 0.6),
        mwpf.HyperEdge([0, 1], 0.7),
        mwpf.HyperEdge([1, 2], 0.8),
        mwpf.HyperEdge([0, 1, 2], 0.3),  # hyper edge
    ]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    visualizer = mwpf.Visualizer(positions=circle_positions(vertex_num))
    syndrome = mwpf.SyndromePattern([0, 1, 2])
    solver.solve(syndrome, visualizer)
    subgraph, bound = solver.subgraph_range(visualizer)
    print(subgraph, bound)
    assert bound.lower == bound.upper
    assert bound.lower.float() == 0.3
    visualizer.save_html(os.path.join(os.path.dirname(__file__), f"test_corner_1.html"))


def test_corner_2():
    """
    test empty edge set
    """
    vertex_num = 4
    weighted_edges = []
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    visualizer = mwpf.Visualizer(positions=circle_positions(vertex_num))
    syndrome = mwpf.SyndromePattern([])
    solver.solve(syndrome, visualizer)
    subgraph, bound = solver.subgraph_range(visualizer)
    print(subgraph, bound)
    assert bound.lower == bound.upper
    assert bound.lower.float() == 0.0
    visualizer.save_html(os.path.join(os.path.dirname(__file__), f"test_corner_2.html"))


def test_corner_3():
    """
    test empty graph
    """
    vertex_num = 0
    weighted_edges = []
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    visualizer = mwpf.Visualizer(positions=circle_positions(vertex_num))
    syndrome = mwpf.SyndromePattern([])
    solver.solve(syndrome, visualizer)
    subgraph, bound = solver.subgraph_range(visualizer)
    print(subgraph, bound)
    assert bound.lower == bound.upper
    assert bound.lower.float() == 0.0
    with open(os.path.join(os.path.dirname(__file__), f"test_corner_2.html"), "w") as f:
        f.write(visualizer.generate_html())


def test_corner_4():
    """
    test duplicate edge set
    """
    vertex_num = 4
    weighted_edges = [
        mwpf.HyperEdge([0], 0.6),
        mwpf.HyperEdge([0, 1], 0.7),
        mwpf.HyperEdge([1, 2], 0.8),
        mwpf.HyperEdge([1, 2], 0.8),
        mwpf.HyperEdge([1, 2], 1.2),
        mwpf.HyperEdge([1, 2], 0.8),
        mwpf.HyperEdge([0, 1, 2], 0.3),  # hyper edge
    ]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    visualizer = mwpf.Visualizer(positions=circle_positions(vertex_num))
    syndrome = mwpf.SyndromePattern([])
    solver.solve(syndrome, visualizer)
    subgraph, bound = solver.subgraph_range(visualizer)
    print(subgraph, bound)
    assert bound.lower == bound.upper
    assert bound.lower.float() == 0.0
    with open(os.path.join(os.path.dirname(__file__), f"test_corner_2.html"), "w") as f:
        f.write(visualizer.generate_html())



from common import *


def test_fp_1():
    vertex_num = 4
    weighted_edges = [
        mwpf.HyperEdge([0], 0.6),
        mwpf.HyperEdge([0, 1], 0.7),
        mwpf.HyperEdge([1, 2], 0.8),
        mwpf.HyperEdge([2, 3], 0.9),
        mwpf.HyperEdge([3], 1.1),
        mwpf.HyperEdge([0, 1, 2], 0.3),  # hyper edge
    ]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    visualizer = mwpf.Visualizer(positions=circle_positions(vertex_num))
    syndrome = mwpf.SyndromePattern([0, 1, 2])
    solver.solve(syndrome, visualizer)
    subgraph, bound = solver.subgraph_range(visualizer)
    print(subgraph, bound)
    assert bound.lower == bound.upper
    assert bound.lower.float() == 0.3
    visualizer.save_html(os.path.join(os.path.dirname(__file__), f"test_fp_1.html"))


def test_fp_2():
    vertex_num = 4
    weighted_edges = [
        mwpf.HyperEdge([0], -0.6),
        mwpf.HyperEdge([0, 1], 0.7),
        mwpf.HyperEdge([1, 2], 0.8),
        mwpf.HyperEdge([2, 3], 0.9),
        mwpf.HyperEdge([3], 1.1),
        mwpf.HyperEdge([0, 1, 2], 0.3),  # hyper edge
    ]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    visualizer = mwpf.Visualizer(positions=circle_positions(vertex_num))
    syndrome = mwpf.SyndromePattern([0, 1, 2])
    solver.solve(syndrome, visualizer)
    subgraph, bound = solver.subgraph_range(visualizer)
    print("subgraph edges:", [edge_index for edge_index in subgraph])
    assert subgraph == [0, 2]
    visualizer.snapshot("original problem", initializer, syndrome, subgraph)
    print(subgraph, bound)
    assert math.isclose(bound.lower.float(), 0.2)
    assert math.isclose(bound.upper.float(), 0.2)
    visualizer.save_html(os.path.join(os.path.dirname(__file__), f"test_fp_2.html"))



"""
Regular Detector Error Model (DEM) does not contain the information of the heralded error.
This HeraldedDetectorErrorModel class provides additional information on the heralded errors.
It is capable of reading bits from the detector which corresponds to the heralded error indicator.

Note that in order to let the tool read a heralded error in the circuit, it is required that 
the heralded error is detected using `DETECTOR rec[...]` where `rec[...]` corresponds to the heralded event.
To help user, we provide a function that automatically adds such detections.

[warning] not all circuits can be used in this class. Notably, there are several requirements:
1. the measurement of a heralded error must be either not detected or uniquely detected by a detector
``
"""

import stim
from .ref_circuit import (
    RefCircuit,
    RefInstruction,
    RefRec,
    RefDetector,
    RefDetectorErrorModel,
    probability_to_weight,
    Predictor,
)
from typing import Sequence
import functools
from dataclasses import dataclass
from frozendict import frozendict
import mwpf_incr
import numpy as np


DEM_MIN_PROBABILITY = 1e-15  # below this value, DEM starts to ignore the error rate


# avoid non-zero small probability to be ignored by the DEM
def dem_probability(probability: float) -> float:
    if probability == 0:
        return 0.0
    return max(DEM_MIN_PROBABILITY, probability)


@dataclass(frozen=True)
class HeraldedDetectorErrorModel:
    ref_circuit: RefCircuit
    # We assume that the detector error model has a non-zero false positive rate to make the decoding
    # graph generation simpler. By default the value is 1e-300
    false_positive_rate: float = DEM_MIN_PROBABILITY

    def __post_init__(self) -> None:
        self.sanity_check()

    def of(
        circuit: stim.Circuit,
        false_positive_rate: float = DEM_MIN_PROBABILITY,
    ) -> "HeraldedDetectorErrorModel":
        return HeraldedDetectorErrorModel(
            ref_circuit=RefCircuit.of(circuit), false_positive_rate=false_positive_rate
        )

    def sanity_check(self) -> None:
        # check basic type
        assert isinstance(
            self.ref_circuit, RefCircuit
        ), "ref_circuit must be a RefCircuit"
        assert isinstance(
            self.false_positive_rate, float
        ), "false_positive_rate must be a float"
        # each heralded error is either not detected or uniquely detected
        for heralded_rec in self.heralded_measurements:
            assert (
                len(self.ref_circuit.rec_to_detectors[heralded_rec]) <= 1
            ), f"abs[{heralded_rec.abs_index(self.ref_circuit)}] is detected by multiple detectors"
            for detector in self.ref_circuit.rec_to_detectors[heralded_rec]:
                assert len(detector.targets) == 1, (
                    f"detector[{self.ref_circuit.detector_to_index[detector]}] detects multiple recs; "
                    + "we require that the detector of a heralded error must only detect one rec"
                )

    @functools.cached_property
    def heralded_instructions(self) -> tuple[RefInstruction, ...]:
        return tuple(
            instruction
            for instruction in self.ref_circuit
            if is_heralded_error(instruction)
        )

    @functools.cached_property
    def heralded_measurements(self) -> tuple[RefRec, ...]:
        return tuple(
            rec
            for instruction in self.ref_circuit
            if is_heralded_error(instruction)
            for rec in instruction.recs
        )

    @functools.cached_property
    def detected_heralded_measurements(self) -> tuple[RefRec, ...]:
        return tuple(
            rec
            for rec in self.heralded_measurements
            if self.ref_circuit.rec_to_detectors[rec]
        )

    @functools.cached_property
    def undetected_heralded_measurements(self) -> tuple[RefRec, ...]:
        return tuple(
            rec
            for rec in self.heralded_measurements
            if not self.ref_circuit.rec_to_detectors[rec]
        )

    @functools.cached_property
    def heralded_detectors(self) -> tuple[RefDetector | None, ...]:
        heralded_measurements = frozenset(self.heralded_measurements)
        return tuple(
            detector if (set(detector.targets) & heralded_measurements) else None
            for detector in self.ref_circuit.detectors
        )

    @functools.cached_property
    def heralded_detector_indices(self) -> tuple[int, ...]:
        return tuple(
            {
                detector_id
                for detector_id, detector in enumerate(self.heralded_detectors)
                if detector is not None
            }
        )

    @functools.cached_property
    def detector_id_to_herald_id(self) -> frozendict[int, int]:
        return frozendict(
            {
                detector_id: herald_id
                for herald_id, detector_id in enumerate(self.heralded_detector_indices)
            }
        )

    @functools.cached_property
    def num_heralds(self) -> int:
        return len(self.heralded_detector_indices)

    @functools.cached_property
    def skeleton_circuit(self) -> RefCircuit:
        """
        The skeleton circuit is a circuit where all the heralded errors are not triggered.
        There will be still some false positive rate remaining to make sure all the possible
        hyperedges still exist in the decoding hypergraph.
        """
        new_instructions = list(self.ref_circuit)
        deleting_indices: list[int] = []
        # first change instruction:
        #     HERALDED_ERASE -> DEPOLARIZE1(false_positive_rate)
        #     HERALDED_PAULI_CHANNEL_1 -> PAULI_CHANNEL_1(...)
        for instruction in self.heralded_instructions:
            instruction_index = self.ref_circuit.instruction_to_index[instruction]
            noise_instruction = heralded_instruction_to_noise_instruction(instruction)
            if noise_instruction is None:
                deleting_indices.append(instruction_index)
                continue
            tiny_noise_instruction = RefInstruction(
                name=noise_instruction.name,
                targets=noise_instruction.targets,
                gate_args=tuple(
                    dem_probability(p * self.false_positive_rate)
                    for p in noise_instruction.gate_args
                ),
            )
            new_instructions[instruction_index] = tiny_noise_instruction
        # then delete detectors of the heralded errors
        for detector in self.heralded_detectors:
            if detector is not None:
                deleting_indices.append(self.ref_circuit.instruction_to_index[detector])
        assert len(set(deleting_indices)) == len(deleting_indices), "bug: duplicate"
        for index in sorted(deleting_indices, reverse=True):
            del new_instructions[index]
        return RefCircuit.of(new_instructions)

    @functools.cached_property
    def skeleton_dem(self) -> RefDetectorErrorModel:
        """
        construct a dem whose detector id corresponds to the detectors of the original circuit
        instead of the skeleton circuit
        """
        ref_dem = self.skeleton_circuit.ref_dem
        # let the dem refer to self.ref_circuit instead
        return RefDetectorErrorModel(
            instructions=ref_dem.instructions, ref_circuit=self.ref_circuit
        )

    @functools.cached_property
    def heralded_dems(
        self,
    ) -> frozendict[RefDetector, RefDetectorErrorModel]:
        ref_dems: dict[RefDetector, RefDetectorErrorModel] = {}
        for detector in self.heralded_detectors:
            if detector is None:
                continue
            ref_rec = detector.targets[0]
            heralded_instruction = ref_rec.instruction
            all_noise_instruction = heralded_instruction_to_noise_instruction(
                heralded_instruction
            )
            if all_noise_instruction is None:
                continue
            assert isinstance(ref_rec, RefRec)
            # remove all the noise channels except the heralded error
            circuit_no_noise = self.ref_circuit.remove_noise_channels(
                keeping={ref_rec.instruction}
            )
            new_circuit_instructions = list(circuit_no_noise)
            # change the heralded error to the noise channel when it happens
            assert (
                len(all_noise_instruction.targets)
                == heralded_instruction.num_measurements
            ), "the following code assumes target has a heralding measurement"
            new_circuit_instructions[
                circuit_no_noise.instruction_to_index[heralded_instruction]
            ] = RefInstruction(
                name=all_noise_instruction.name,
                targets=(all_noise_instruction.targets[ref_rec.bias],),
                gate_args=all_noise_instruction.gate_args,
            )
            new_circuit = RefCircuit.of(new_circuit_instructions)
            heralded_dem = RefDetectorErrorModel(
                instructions=new_circuit.ref_dem.instructions,
                ref_circuit=self.ref_circuit,
            )
            if not heralded_dem.hyperedges:
                # if there is no hyperedge, we don't need this detector at all
                continue
            for hyperedge in heralded_dem.hyperedges:
                assert (
                    hyperedge.detectors in self.skeleton_dem.hyperedges_detectors_set
                ), (
                    "bug: the skeleton graph doesn't have the hyperedge, "
                    + "this might causes issue when constructing decoders"
                )
            ref_dems[detector] = heralded_dem

        return frozendict(ref_dems)

    def __str__(self) -> str:
        result = "HeraldedDetectorErrorModel:"
        result += "\n    skeleton hypergraph:"
        for dem_hyperedge in self.skeleton_dem.hyperedges:
            result += (
                f"\n        {', '.join([f'D{v}' for v in sorted(dem_hyperedge.detectors)])}: {dem_hyperedge.probability:.6e}"
                + f" ({', '.join([f'L{v}' for v in sorted(dem_hyperedge.observables)])})"
            )
        for detector, ref_dem in self.heralded_dems.items():
            result += f"\n    heralded hypergraph on D{self.ref_circuit.detector_to_index[detector]}:"
            for hyperedge in ref_dem.hyperedges:
                result += (
                    f"\n        {', '.join([f'D{v}' for v in sorted(hyperedge.detectors)])}: {hyperedge.probability:.6e}"
                    + f" ({', '.join([f'L{v}' for v in sorted(hyperedge.observables)])})"
                )
        return result

    @functools.cached_property
    def hyperedge_to_index(self) -> frozendict[frozenset[int], int]:
        return frozendict(
            {
                dem_hyperedge.detectors: edge_index
                for edge_index, dem_hyperedge in enumerate(self.skeleton_dem.hyperedges)
            }
        )

    @functools.cached_property
    def herald_fault_map(self) -> tuple[frozendict[int, tuple[float, int]], ...]:
        heralds: list[frozendict[int, tuple[float, int]]] = []
        for detector_id in self.heralded_detector_indices:
            detector = self.heralded_detectors[detector_id]
            assert detector is not None
            if detector not in self.heralded_dems:
                heralds.append(frozendict({}))
                continue
            sub_dem = self.heralded_dems[detector]
            heralds.append(
                frozendict(
                    {
                        self.hyperedge_to_index[hyperedge.detectors]: (
                            hyperedge.probability,
                            sum(1 << k for k in hyperedge.observables),
                        )
                        for hyperedge in sub_dem.hyperedges
                    }
                )
            )
        return tuple(heralds)

    @functools.cached_property
    def initializer(self) -> mwpf_incr.SolverInitializer:
        vertex_num = self.skeleton_dem._dem.num_detectors
        weighted_edges = [
            mwpf_incr.HyperEdge(
                dem_hyperedge.detectors,
                probability_to_weight(dem_hyperedge.probability),
            )
            for dem_hyperedge in self.skeleton_dem.hyperedges
        ]
        heralds = [
            {edge_index: probability_to_weight(p) for edge_index, (p, _) in dic.items()}
            for dic in self.herald_fault_map
        ]
        return mwpf_incr.SolverInitializer(vertex_num, weighted_edges, heralds=heralds)

    @functools.cached_property
    def predictor(self) -> "HeraldedDemPredictor":
        fault_masks_with_p = tuple(
            (sum(1 << k for k in dem_hyperedge.observables), dem_hyperedge.probability)
            for dem_hyperedge in self.skeleton_dem.hyperedges
        )
        herald_detectors = frozenset(
            {
                detector_id
                for detector_id, detector in enumerate(self.heralded_detectors)
                if detector is not None
            }
        )
        return HeraldedDemPredictor(
            fault_masks_with_p=fault_masks_with_p,
            herald_detectors=herald_detectors,
            detector_id_to_herald_id=self.detector_id_to_herald_id,
            herald_fault_map=self.herald_fault_map,
            num_dets=self.skeleton_dem._dem.num_detectors,
            num_obs=self.skeleton_dem._dem.num_observables,
        )


@dataclass(frozen=True)
class HeraldedDemPredictor(Predictor):
    """
    the correction should be chosen based on the heralded error: if certain observable achieves higher
    probability, we should choose the correction based on that observable; this is a dynamic behavior
    ideally, we should not spend too much computation in Python.
    How about given the subgraph object and then calculate the prediction? The subgraph should be pretty sparse
    """

    fault_masks_with_p: tuple[tuple[int, float], ...]
    herald_detectors: frozenset[int]
    detector_id_to_herald_id: frozendict[int, int]
    herald_fault_map: tuple[frozendict[int, tuple[float, int]], ...]
    num_dets: int
    num_obs: int

    def syndrome_of(self, dets_bit_packed: np.ndarray) -> mwpf_incr.SyndromePattern:
        detectors: set[int] = set(
            np.flatnonzero(
                np.unpackbits(dets_bit_packed, count=self.num_dets, bitorder="little")
            )
        )
        # the heralded detectors are not passed to the decoder
        defect_vertices = detectors - self.herald_detectors
        # instead, they are passed as heralds
        heralds = [
            self.detector_id_to_herald_id[detector_id]
            for detector_id in detectors & self.herald_detectors
        ]
        return mwpf_incr.SyndromePattern(defect_vertices=defect_vertices, heralds=heralds)

    def prediction_of(
        self, syndrome: mwpf_incr.SyndromePattern, subgraph: Sequence[int]
    ) -> int:
        prediction: int = 0
        heralds: list[int] = syndrome.heralds
        for edge_index in subgraph:
            fault_mask, p = self.fault_masks_with_p[edge_index]
            # also iterate over the heralded errors to see if there is a more probable logical correction
            for herald_id in heralds:
                if edge_index in self.herald_fault_map[herald_id]:
                    new_p, new_fault_mask = self.herald_fault_map[herald_id][edge_index]
                    if new_p > p:
                        p = new_p
                        fault_mask = new_fault_mask
            prediction ^= fault_mask
        return prediction

    def num_detectors(self) -> int:
        return self.num_dets

    def num_observables(self) -> int:
        return self.num_obs

    @property
    def is_dynamic(self) -> bool:
        return len(self.herald_detectors) > 0


def add_herald_detectors(circuit: stim.Circuit) -> stim.Circuit:
    """
    Add detectors for heralded errors to the circuit, if they do not present in the original circuit.
    Note that the circuit will be fully expanded after this function. The heralded detectors added
    will be following each heralded error instruction.
    """
    ref_circuit = RefCircuit.of(circuit)
    heralded_instructions = tuple(
        instruction for instruction in ref_circuit if is_heralded_error(instruction)
    )
    new_instructions = list(ref_circuit)
    for heralded_instruction in reversed(heralded_instructions):
        for rec in reversed(heralded_instruction.recs):
            if not ref_circuit.rec_to_detectors[rec]:
                new_instructions.insert(
                    ref_circuit.instruction_to_index[heralded_instruction] + 1,
                    RefInstruction(
                        name="DETECTOR",
                        targets=(rec,),
                    ),
                )
    return RefCircuit.of(new_instructions).circuit()


def remove_herald_detectors(circuit: stim.Circuit) -> stim.Circuit:
    """
    Remove all detectors of the heralded errors. It will only remove detectors that uniquely detects the heralded errors,
    and panic if some composite detectors involve the heralded error measurement.
    """
    ref_circuit = RefCircuit.of(circuit)
    new_instructions = []
    for instruction in ref_circuit:
        if instruction.name != "DETECTOR":
            new_instructions.append(instruction)
            continue
        assert len(instruction.targets) > 0, "bug: detector without target"
        if len(instruction.targets) > 1:
            # check that none of these recs corresponds to an heralded instruction
            for rec in instruction.targets:
                assert isinstance(rec, RefRec)
                assert not is_heralded_error(rec.instruction), ""
            new_instructions.append(instruction)
            continue
        rec = instruction.targets[0]
        assert isinstance(rec, RefRec)
        if not is_heralded_error(rec.instruction):
            new_instructions.append(instruction)
    return RefCircuit.of(new_instructions).circuit()


def is_heralded_error(instruction: RefInstruction) -> bool:
    if instruction.name in ["HERALDED_ERASE", "HERALDED_PAULI_CHANNEL_1"]:
        return True
    if "HERALDED" in instruction.name:
        raise NotImplementedError(
            f"Instruction {instruction.name} has 'HERALDED' in its name but is not implemented yet."
        )
    return False


def heralded_instruction_to_noise_instruction(
    instruction: RefInstruction,
) -> RefInstruction | None:
    if instruction.name == "HERALDED_ERASE":
        heralded_probability = instruction.gate_args[0]
        if heralded_probability == 0:
            return None
        return RefInstruction(
            name="DEPOLARIZE1",
            targets=instruction.targets,
            gate_args=(0.75,),
        )
    elif instruction.name == "HERALDED_PAULI_CHANNEL_1":
        pI, pX, pY, pZ = instruction.gate_args
        if pX + pY + pZ == 0:
            return None
        p_sum = pI + pX + pY + pZ
        return RefInstruction(
            name="PAULI_CHANNEL_1",
            targets=instruction.targets,
            gate_args=(pX / p_sum, pY / p_sum, pZ / p_sum),
        )
    return None



from common import *
import subprocess
import os
import sys


# code = mwpf.CodeCapacityRepetitionCode(d=3, p=0.01)
# code = mwpf.CodeCapacityPlanarCode(d=3, p=0.01)
code = mwpf.CodeCapacityTailoredCode(d=5, pxy=0.001, pz=0.1)
# code = mwpf.CodeCapacityColorCode(d=3, p=0.01)

# either randomly generate errors
# code.generate_random_errors()

# or manually set errors
# code.set_physical_errors([1, 2, 3, 4, 5, 6])

# or manually set defect
code.set_defect_vertices([10, 11, 16, 17])

initializer = code.get_initializer()

# solver configuration
config = {"cluster_node_limit": 100, "timeout": 10.0}
solver = mwpf.SolverSerialJointSingleHair(initializer, config)

"""
run the solver
"""

git_root_dir = (
    subprocess.run(
        "git rev-parse --show-toplevel",
        cwd=os.path.dirname(os.path.abspath(__file__)),
        shell=True,
        check=True,
        capture_output=True,
    )
    .stdout.decode(sys.stdout.encoding)
    .strip(" \r\n")
)
data_folder = os.path.join(git_root_dir, "visualize", "data")

visualizer = mwpf.Visualizer(positions=code.get_positions())
visualizer.snapshot("syndrome", code)

syndrome = code.get_syndrome()
print(f"syndrome: {syndrome}")
solver.solve(syndrome, visualizer)
subgraph, bound = solver.subgraph_range(visualizer)

print(f"subgraph: {subgraph}")
print(f"bound: {(bound.lower, bound.upper)}")
if bound.lower == bound.upper:
    print("[success] optimal! 🤩")
else:
    print("[potential failure] may be suboptimal... 😅")

visualizer.save_html(
    os.path.join(os.path.dirname(__file__), f"python_test_visualize.html")
)



from common import *
import pickle


def test_basic_pickle():
    vertex_num = 2
    weighted_edges = [
        mwpf.HyperEdge([0, 1], 100),
    ]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)

    pickled = pickle.dumps(solver)
    config = solver.config
    print(config)
    print(solver)
    print(pickled)

    print("\n\n#### pickle and then unpickle  ####\n\n")

    solver2 = pickle.loads(pickled)
    config2 = solver2.config
    print(config2)
    print(solver2)

    assert config == config2



import os, math, pytest, sys

""" force import either mwpf or mwpf_rational """
# import mwpf
# import mwpf_rational as mwpf

""" automatic import based on which is available """
if "mwpf" not in globals():
    try:
        import mwpf
    except ImportError as e:
        print(e)
        print("mwpf package not available, trying others...")

if "mwpf" not in globals():
    try:
        import mwpf_rational as mwpf
    except ImportError as e:
        print(e)
        print("mwpf_rational package not available, trying others...")

if "mwpf" not in globals():
    try:
        import mwpf_incr as mwpf
    except ImportError as e:
        print(e)
        print("mwpf_incr package not available, trying others...")

if "mwpf" not in globals():
    try:
        import mwpf_fast as mwpf
    except ImportError as e:
        print(e)
        print("mwpf_fast package not available, trying others...")


def circle_positions(n: int):
    positions = []
    for i in range(n):
        positions.append(
            mwpf.VisualizePosition(
                0.5 + 0.5 * math.cos(2 * math.pi * i / n),
                0.5 + 0.5 * math.sin(2 * math.pi * i / n),
                0,
            )
        )
    return positions



from common import *
import traceback


def test_basic_panic():
    vertex_num = 2
    weighted_edges = [
        mwpf.HyperEdge([0, 1], 100),
    ]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)

    visualizer = mwpf.Visualizer(
        positions=[mwpf.VisualizePosition(0, 0, 0), mwpf.VisualizePosition(1, 0, 0)]
    )
    syndrome = mwpf.SyndromePattern([0])
    try:
        try:
            solver.solve(syndrome, visualizer)  # unsolvable, and should panic

            visualizer.save_html(
                os.path.join(os.path.dirname(__file__), f"test_basic_panic.html")
            )

            solver.subgraph()
        except BaseException as panic:
            raise ValueError(mwpf.panic_text_of(solver, syndrome)) from panic
    except BaseException:
        panic_text = traceback.format_exc()
        # print(panic_text)
        assert "######## MWPF Sinter Decoder Panic ########" in panic_text
        pass
    else:
        assert False, "panic expected"

    # after the panic, the solver should still be able to report the information
    post_panic_initializer = solver.get_initializer()
    assert post_panic_initializer.to_json() == initializer.to_json()
    post_panic_config = solver.config
    print(post_panic_config)



import math
import pathlib
from typing import Tuple, Any, Optional
import mwpf_incr
from mwpf_incr import (  # type: ignore
    SyndromePattern,
    HyperEdge,
    SolverInitializer,
    Solver,
    BP,
    BenchmarkSuite,
    WeightRange,
)
from dataclasses import dataclass, field
import pickle
import json
import traceback
from enum import Enum
import random
import numpy as np
import stim
from io import BufferedReader, BufferedWriter
import time
import struct
from contextlib import nullcontext
from .ref_circuit import *
from .heralded_dem import *

available_decoders = [
    "Solver",  # the solver with the highest accuracy, but may change across different versions
    "SolverSerialJointSingleHair",
    "SolverSerialSingleHair",
    "SolverSerialUnionFind",
]

default_cluster_node_limit: int = 50


@dataclass
class DecoderPanic:
    initializer: SolverInitializer
    config: dict
    syndrome: SyndromePattern
    panic_message: str


class PanicAction(Enum):
    RAISE = 1  # raise the panic with proper message to help debugging
    CATCH = 2  # proceed with normal decoding and return all-0 result


@dataclass
class SinterMWPFIncrDecoder:
    """
    Use MWPF to predict observables from detection events.

    Args:
        decoder_type: decoder class used to construct the MWPF decoder.  in the Rust implementation, all of them inherits from the class of `SolverSerialPlugins` but just provide different plugins for optimizing the primal and/or dual solutions. For example, `SolverSerialUnionFind` is the most basic solver without any plugin: it only grows the clusters until the first valid solution appears; some more optimized solvers uses one or more plugins to further optimize the solution, which requires longer decoding time.

        cluster_node_limit (alias: c): The maximum number of nodes in a cluster, used to tune the performance of the decoder. The default value is 50.
    """

    decoder_type: str = "SolverSerialJointSingleHair"
    cluster_node_limit: Optional[int] = None
    c: Optional[int] = None  # alias of `cluster_node_limit`, will override it
    timeout: Optional[float] = None
    with_progress: bool = False
    circuit: Optional[stim.Circuit] = None  # RefCircuit is not picklable
    # this parameter itself doesn't do anything to load the circuit but only check whether the circuit is indeed loaded
    pass_circuit: bool = False

    # record panic data and controls whether the raise the panic or simply record them
    panic_action: PanicAction = PanicAction.CATCH
    panic_cases: list[DecoderPanic] = field(default_factory=list)

    # record benchmark suite when enabled
    benchmark_suite_filename: Optional[str] = None
    # record decoding data when enabled, including decoding time, primal and dual weights, etc.
    trace_filename: Optional[str] = None

    # adding BP decoder as a pre-decoder
    bp: bool = False
    max_iter: int = 0  # by default "adaptive"
    bp_method: str = "ms"  # 'product_sum' or 'minimum_sum'
    ms_scaling_factor: float = 0.625  # usually better than the original default of 1.0
    schedule: str = "parallel"  #  'parallel', 'serial', or 'serial_relative'
    omp_thread_count: int = 1
    random_schedule_seed: int = 0
    serial_schedule_order: Optional[list[int]] = None
    bp_weight_mix_ratio: float = 1.0
    floor_weight: Optional[float] = (
        None  # when updating the mwpf weights, all the weights are enforced to be no less than this value; by default 0 which enforces all the weights to be non-negative
    )
    # by default BP may converge and directly return the result;
    # sometimes it's better if BP cannot directly return and always use the post-processing to decode
    bp_converge: bool = True

    @property
    def _cluster_node_limit(self) -> int:
        if self.cluster_node_limit is not None:
            assert self.c is None, "Cannot set both `cluster_node_limit` and `c`."
            return self.cluster_node_limit
        elif self.c is not None:
            assert (
                self.cluster_node_limit is None
            ), "Cannot set both `cluster_node_limit` and `c`."
            return self.c
        return default_cluster_node_limit

    @property
    def config(self) -> dict[str, Any]:
        return dict(cluster_node_limit=self._cluster_node_limit)

    def with_circuit(self, circuit: stim.Circuit | None) -> "SinterMWPFIncrDecoder":
        if circuit is None:
            self.circuit = None
            return self
        assert isinstance(circuit, stim.Circuit)
        self.circuit = circuit.copy()
        return self

    def common_prepare(
        self, dem: "stim.DetectorErrorModel"
    ) -> tuple[Any, Predictor, Any]:
        if self.pass_circuit:
            assert (
                self.circuit is not None
            ), "The circuit is not loaded but the flag `pass_circuit` is True"

        solver, predictor = construct_decoder_and_predictor(
            dem,
            decoder_type=self.decoder_type,
            config=self.config,
            ref_circuit=(
                RefCircuit.of(self.circuit) if self.circuit is not None else None
            ),
        )
        assert (
            dem.num_detectors == predictor.num_detectors()
        ), "Mismatched number of detectors, are you using the corresponding circuit of dem?"
        assert (
            dem.num_observables == predictor.num_observables()
        ), "Mismatched number of observables, are you using the corresponding circuit of dem?"

        # construct bp decoder if requested
        bp_decoder: Optional[Any] = None
        if self.bp:
            if self.circuit is not None:
                assert (
                    not predictor.is_dynamic
                ), "BP is not supported for dynamic predictors, e.g., in presence of heralded errors."

            from ldpc import BpDecoder
            from ldpc.ckt_noise.dem_matrices import (
                detector_error_model_to_check_matrices,
            )

            bp_matrices = detector_error_model_to_check_matrices(
                dem, allow_undecomposed_hyperedges=True
            )
            bp_decoder = BpDecoder(
                pcm=bp_matrices.check_matrix,
                error_channel=list(bp_matrices.priors),
                max_iter=self.max_iter,
                bp_method=self.bp_method,
                ms_scaling_factor=self.ms_scaling_factor,
                schedule=self.schedule,
                omp_thread_count=self.omp_thread_count,
                serial_schedule_order=self.serial_schedule_order,
                input_vector_type="syndrome",
            )

        return solver, predictor, bp_decoder

    def compile_decoder_for_dem(
        self,
        *,
        dem: "stim.DetectorErrorModel",
    ) -> "MwpfIncrCompiledDecoder":
        solver, predictor, bp_decoder = self.common_prepare(dem)

        benchmark_suite: Optional[BenchmarkSuite] = None
        if self.benchmark_suite_filename is not None:
            benchmark_suite = BenchmarkSuite(solver.get_initializer())
            solver = None

        return MwpfIncrCompiledDecoder(
            solver,
            predictor,
            dem.num_detectors,
            dem.num_observables,
            panic_action=self.panic_action,
            panic_cases=self.panic_cases,  # record all the panic information to the same place
            benchmark_suite=benchmark_suite,
            benchmark_suite_filename=self.benchmark_suite_filename,
            trace_filename=self.trace_filename,
            bp_decoder=bp_decoder,
            bp_weight_mix_ratio=self.bp_weight_mix_ratio,
            floor_weight=self.floor_weight,
            bp_converge=self.bp_converge,
        )

    def decode_via_files(
        self,
        *,
        num_shots: int,
        num_dets: int,
        num_obs: int,
        dem_path: pathlib.Path,
        dets_b8_in_path: pathlib.Path,
        obs_predictions_b8_out_path: pathlib.Path,
        tmp_dir: pathlib.Path,
    ) -> None:
        dem = stim.DetectorErrorModel.from_file(dem_path)

        solver, predictor, bp_decoder = self.common_prepare(dem)

        assert num_dets == predictor.num_detectors()
        assert num_obs == predictor.num_observables()

        benchmark_suite: Optional[BenchmarkSuite] = None
        if self.benchmark_suite_filename is not None:
            benchmark_suite = BenchmarkSuite(solver.get_initializer())
            solver = None

        num_det_bytes = math.ceil(num_dets / 8)
        with (
            open(self.trace_filename, "wb")
            if self.trace_filename is not None
            else nullcontext()
        ) as trace_f:
            with open(dets_b8_in_path, "rb") as dets_in_f:
                with open(obs_predictions_b8_out_path, "wb") as obs_out_f:
                    for dets_bit_packed in iter_det(
                        dets_in_f, num_shots, num_det_bytes, self.with_progress
                    ):
                        prediction = decode_common(
                            dets_bit_packed=dets_bit_packed,
                            predictor=predictor,
                            solver=solver,
                            num_dets=num_dets,
                            num_obs=num_obs,
                            panic_action=self.panic_action,
                            panic_cases=self.panic_cases,
                            benchmark_suite=benchmark_suite,
                            trace_f=trace_f,
                            bp_decoder=bp_decoder,
                            bp_weight_mix_ratio=self.bp_weight_mix_ratio,
                            floor_weight=self.floor_weight,
                            bp_converge=self.bp_converge,
                        )
                        obs_out_f.write(
                            int(prediction).to_bytes(
                                (num_obs + 7) // 8, byteorder="little"
                            )
                        )

        if benchmark_suite is not None:
            benchmark_suite.save_cbor(self.benchmark_suite_filename)


def iter_det(
    f: BufferedReader,
    num_shots: int,
    num_det_bytes: int,
    with_progress: bool = False,
) -> Iterable[np.ndarray]:
    if with_progress:
        from tqdm import tqdm

        pbar = tqdm(total=num_shots, desc="shots")
    for _ in range(num_shots):
        if with_progress:
            pbar.update(1)
        dets_bit_packed = np.fromfile(f, dtype=np.uint8, count=num_det_bytes)
        if dets_bit_packed.shape != (num_det_bytes,):
            raise IOError("Missing dets data.")
        yield dets_bit_packed


def construct_decoder_and_predictor(
    model: "stim.DetectorErrorModel",
    decoder_type: Any,
    config: dict[str, Any],
    ref_circuit: Optional[RefCircuit] = None,
) -> Tuple[Any, Predictor]:

    if ref_circuit is not None:
        heralded_dem = HeraldedDetectorErrorModel(ref_circuit=ref_circuit)
        initializer = heralded_dem.initializer
        predictor: Predictor = heralded_dem.predictor
    else:
        ref_dem = RefDetectorErrorModel.of(dem=model)
        initializer = ref_dem.initializer
        predictor = ref_dem.predictor

    if decoder_type is None:
        # default to the solver with highest accuracy
        decoder_cls = Solver
    elif isinstance(decoder_type, str):
        decoder_cls = getattr(mwpf_incr, decoder_type)
    else:
        decoder_cls = decoder_cls
    return (
        decoder_cls(initializer, config=config),
        predictor,
    )


def panic_text_of(solver, syndrome) -> str:
    initializer = solver.get_initializer()
    config = solver.config
    syndrome
    panic_text = f"""
######## MWPF Sinter Decoder Panic ######## 
solver_initializer: dict = json.loads('{initializer.to_json()}')
config: dict = json.loads('{json.dumps(config)}')
syndrome: dict = json.loads('{syndrome.to_json()}')
######## PICKLE DATA ######## 
solver_initializer: SolverInitializer = pickle.loads({pickle.dumps(initializer)!r})
config: dict = pickle.loads({pickle.dumps(config)!r})
syndrome: SyndromePattern = pickle.loads({pickle.dumps(syndrome)!r})
######## End Panic Information ######## 
"""
    return panic_text


@dataclass
class SinterHUFIncrDecoder(SinterMWPFIncrDecoder):
    decoder_type: str = "SolverSerialUnionFind"
    cluster_node_limit: int = 0


@dataclass
class SinterSingleHairIncrDecoder(SinterMWPFIncrDecoder):
    decoder_type: str = "SolverSerialSingleHair"
    cluster_node_limit: int = 0


@dataclass
class MwpfIncrCompiledDecoder:
    solver: Any
    predictor: Predictor
    num_dets: int
    num_obs: int
    panic_action: PanicAction
    panic_cases: list[DecoderPanic]
    benchmark_suite: Optional[BenchmarkSuite]
    trace_filename: Optional[str]
    benchmark_suite_filename: Optional[str]
    bp_decoder: Any
    bp_weight_mix_ratio: float
    floor_weight: Optional[float]
    bp_converge: bool = True

    def decode_shots_bit_packed(
        self,
        *,
        bit_packed_detection_event_data: "np.ndarray",
    ) -> "np.ndarray":
        num_shots = bit_packed_detection_event_data.shape[0]
        predictions = np.zeros(
            shape=(num_shots, (self.num_obs + 7) // 8), dtype=np.uint8
        )
        with (
            open(self.trace_filename, "wb")
            if self.trace_filename is not None
            else nullcontext()
        ) as trace_f:
            for shot in range(num_shots):
                dets_bit_packed = bit_packed_detection_event_data[shot]
                prediction = decode_common(
                    dets_bit_packed=dets_bit_packed,
                    predictor=self.predictor,
                    solver=self.solver,
                    num_dets=self.num_dets,
                    num_obs=self.num_obs,
                    panic_action=self.panic_action,
                    panic_cases=self.panic_cases,
                    benchmark_suite=self.benchmark_suite,
                    trace_f=trace_f,
                    bp_decoder=self.bp_decoder,
                    bp_weight_mix_ratio=self.bp_weight_mix_ratio,
                    floor_weight=self.floor_weight,
                    bp_converge=self.bp_converge,
                )
                predictions[shot] = np.packbits(
                    np.array(
                        list(np.binary_repr(prediction, width=self.num_obs))[::-1],
                        dtype=np.uint8,
                    ),
                    bitorder="little",
                )

        if self.benchmark_suite is not None:
            self.benchmark_suite.save_cbor(self.benchmark_suite_filename)

        return predictions


def decode_common(
    dets_bit_packed: np.ndarray,
    predictor: Predictor,
    solver: Any,
    num_dets: int,
    num_obs: int,
    panic_action: PanicAction,
    panic_cases: list[DecoderPanic],
    benchmark_suite: Optional[BenchmarkSuite],
    trace_f: Optional[BufferedWriter],
    bp_decoder: Any,
    bp_weight_mix_ratio: float,
    floor_weight: Optional[float],
    bp_converge: bool = True,
) -> int:
    syndrome = predictor.syndrome_of(dets_bit_packed)
    if solver is None:
        if benchmark_suite is not None:
            benchmark_suite.append(syndrome)
        prediction = 0
    else:
        try:
            start = time.perf_counter()
            if bp_decoder is not None:
                dets_bits = np.unpackbits(
                    dets_bit_packed, count=num_dets, bitorder="little"
                )
                bp_solution = bp_decoder.decode(dets_bits)
                if bp_decoder.converge and bp_converge:
                    prediction = predictor.prediction_of(
                        syndrome, np.flatnonzero(bp_solution)
                    )
                else:
                    syndrome = SyndromePattern(
                        defect_vertices=syndrome.defect_vertices,
                        override_weights=list(bp_decoder.log_prob_ratios),
                        override_ratio=bp_weight_mix_ratio,
                        floor_weight=floor_weight,
                    )
                    solver.solve(syndrome)
                    if trace_f is None:
                        subgraph = solver.subgraph()
                    else:
                        subgraph, bound = solver.subgraph_range()
                        record_trace(trace_f, time.perf_counter() - start, bound)
                    prediction = predictor.prediction_of(syndrome, subgraph)
            else:
                solver.solve(syndrome)
                if trace_f is None:
                    subgraph = solver.subgraph()
                else:
                    subgraph, bound = solver.subgraph_range()
                    record_trace(trace_f, time.perf_counter() - start, bound)
                prediction = predictor.prediction_of(syndrome, subgraph)
        except BaseException as e:
            panic_cases.append(
                DecoderPanic(
                    initializer=solver.get_initializer(),
                    config=solver.config,
                    syndrome=syndrome,
                    panic_message=traceback.format_exc(),
                )
            )
            if "<class 'KeyboardInterrupt'>" in str(e):
                raise e
            elif panic_action == PanicAction.RAISE:
                raise e
                raise ValueError(panic_text_of(self.solver, syndrome)) from e
            elif panic_action == PanicAction.CATCH:
                prediction = random.getrandbits(num_obs)
    return prediction


def record_trace(trace_f: BufferedWriter, elapsed: float, bound: WeightRange):
    trace_f.write(struct.pack("f", elapsed))
    trace_f.write(struct.pack("f", bound.lower.float()))
    trace_f.write(struct.pack("f", bound.upper.float()))
    trace_f.write(struct.pack("f", 0))  # reserved



"""
The Circuit object in stim always write in relative measurement index, 
which is great for writing loops but not so great for analyzing the measurements.
Especially, if we want to analyze the effect of certain heralded errors, we will
need to keep the rest of the measurements in the same place. However, removing or 
adding one heralded error will change all the relative measurement indices, making it
especially hard to track and analyze the circuit.

This class is used to convert the relative measurement index to referenced measurement
 and vice versa. This will allow us to manipulate the circuit by adding or removing
heralded errors more easily.

This is done by adding a custom class `RefRec` in place of the normal `GateTarget`
object returned by the `stim.target_rec(lookback)` function.
In contrast to the negative integer `lookback` parameters, the `bias` integer is a non-negative
index bias over the `RefInstruction` that generates the measurements.
In this way, we are free to manipulate the instructions without worrying that the measurements
indices are messed up.
The `RefCircuit` object will automatically calculate the new relative indices so that the circuit
works properly, and also it will generate a mapping between indices of the old circuit and new circuit.

```python
circuit = stim.Circuit(...)
ref_circuit = RefCircuit.of(circuit)
print(ref_circuit)  # print the circuit in absolute indices
circuit_2 = ref_circuit.to_circuit()  # convert the ref_circuit back to stim.Circuit
print(circuit_2)  # print the circuit in relative indices
```
"""

import stim
from dataclasses import dataclass, field
from typing import Iterator, Iterable, TypeAlias, Collection, Protocol, Sequence, Any
import functools
import numpy as np
from functools import reduce
from frozendict import frozendict
from frozenlist import FrozenList
import mwpf_incr


@dataclass(frozen=True)
class RefRec:
    instruction: "RefInstruction"
    bias: int

    def __post_init__(self):
        if self.bias < 0:
            raise ValueError("The bias must be a non-negative integer.")

    def abs_index(self, circuit: "RefCircuit") -> int:
        return circuit.rec_to_index[self]

    def rel_index(self, circuit: "RefCircuit", instruction: "RefInstruction") -> int:
        """get the relative rec index before executing the instruction"""
        return circuit.rec_to_index[self] - circuit.instruction_to_rec_bias[instruction]

    def __eq__(self, other: object) -> bool:
        return self is other  # avoiding value-based comparison

    def __hash__(self) -> int:
        return hash(id(self))


@dataclass(frozen=True)
class RefInstruction:
    name: str
    targets: tuple[stim.GateTarget | RefRec, ...] = ()
    gate_args: tuple[float, ...] = ()
    tag: str = ""
    recs: FrozenList[RefRec] = field(default_factory=FrozenList)

    @property
    def num_measurements(self) -> int:
        return len(self.recs)

    def index(self, circuit: "RefCircuit") -> int:
        return circuit.instruction_to_index[self]

    def __eq__(self, other: object) -> bool:
        return self is other  # avoiding value-based comparison

    def __hash__(self) -> int:
        return hash(id(self))

    @staticmethod
    def new_heralded_erase(
        targets: tuple[stim.GateTarget | RefRec, ...], p: float
    ) -> "RefInstruction":
        instruction = RefInstruction(
            name="HERALDED_ERASE",
            targets=targets,
            gate_args=(p,),
        )
        for bias, target in enumerate(targets):
            assert isinstance(target, stim.GateTarget)
            instruction.recs.append(
                RefRec(
                    instruction=instruction,
                    bias=bias,
                )
            )
        instruction.recs.freeze()  # do not allow further edit
        return instruction

    @staticmethod
    def new_heralded_pauli_channel_1(
        targets: tuple[stim.GateTarget | RefRec, ...],
        pI: float = 0,
        pX: float = 0,
        pY: float = 0,
        pZ: float = 0,
    ) -> "RefInstruction":
        instruction = RefInstruction(
            name="HERALDED_PAULI_CHANNEL_1",
            targets=targets,
            gate_args=(pI, pX, pY, pZ),
        )
        for bias, target in enumerate(targets):
            assert isinstance(target, stim.GateTarget)
            instruction.recs.append(
                RefRec(
                    instruction=instruction,
                    bias=bias,
                )
            )
        instruction.recs.freeze()  # do not allow further edit
        return instruction

    @staticmethod
    def new_detector(targets: tuple[stim.GateTarget | RefRec, ...]) -> "RefInstruction":
        for target in targets:
            assert isinstance(target, RefRec)
        return RefInstruction(
            name="DETECTOR",
            targets=targets,
        )


RefDetector: TypeAlias = RefInstruction


@dataclass(frozen=True)
class RefCircuit:
    instructions: tuple[RefInstruction, ...]

    def __post_init__(self):
        self.sanity_check()

    @staticmethod
    def of(
        circuit: stim.Circuit | Iterable[RefInstruction] | None = None,
    ) -> "RefCircuit":
        if circuit is None:
            return RefCircuit(instructions=tuple())
        if not isinstance(circuit, stim.Circuit):
            return RefCircuit(instructions=tuple(circuit))
        instructions: list[RefInstruction] = []
        recs: list[RefRec] = []
        for instruction in circuit.flattened():
            # rewrite the gate target to absolute measurement indices
            ref_targets: list[stim.GateTarget | RefRec] = []
            for target in instruction.targets_copy():
                if target.is_measurement_record_target:
                    ref_targets.append(recs[target.value])
                else:
                    ref_targets.append(target)
            # construct the instruction
            ref_instruction = RefInstruction(
                name=instruction.name,
                targets=tuple(ref_targets),
                gate_args=tuple(instruction.gate_args_copy()),
                **extra_instruction_params(instruction),
            )
            instructions.append(ref_instruction)
            # add the measurement to the measurement list
            for bias in range(instruction.num_measurements):
                reference_rec = RefRec(
                    instruction=ref_instruction,
                    bias=bias,
                )
                ref_instruction.recs.append(reference_rec)
                recs.append(reference_rec)
            ref_instruction.recs.freeze()  # do not allow further edit
        return RefCircuit(instructions=tuple(instructions))

    def __repr__(self) -> str:
        repr_str = "RefCircuit: {  # not executable\n"
        repr_str += "\n".join("    " + e for e in str(self).splitlines())
        repr_str += "\n}"
        return repr_str

    def __str__(self) -> str:
        # print the circuit in absolute indices; this helps writing test cases
        code_str = ""
        for instruction, stim_instruction in zip(
            self.instructions, self.stim_instructions
        ):
            rec_count: dict[RefRec, int] = {}
            for ref_target in instruction.targets:
                if isinstance(ref_target, RefRec):
                    if ref_target in rec_count:
                        rec_count[ref_target] += 1
                    else:
                        rec_count[ref_target] = 1
            instruction_str = str(stim_instruction)
            # change all the relative indices to absolute indices
            for ref_rec, count in rec_count.items():
                # this count checking is safe because "rec[..]" becomes "rec[..\C"
                # in the tag parameter, and the character "]" cannot appear elsewhere
                relative_index = ref_rec.rel_index(self, instruction)
                assert instruction_str.count(f"rec[{relative_index}]") == count
                instruction_str = instruction_str.replace(
                    f"rec[{relative_index}]", f"abs[{ref_rec.abs_index(self)}]"
                )
            if len(code_str) > 0:
                code_str += "\n"
            code_str += instruction_str
        return code_str

    def __iter__(self) -> Iterator[RefInstruction]:
        for instruction in self.instructions:
            yield instruction

    @functools.cached_property
    def recs(self) -> tuple[RefRec, ...]:
        recs: list[RefRec] = []
        for instruction in self.instructions:
            recs.extend(instruction.recs)
        return tuple(recs)

    @functools.cached_property
    def rec_to_index(self) -> frozendict[RefRec, int]:
        return frozendict({ref_rec: index for index, ref_rec in enumerate(self.recs)})

    @functools.cached_property
    def rec_to_detectors(self) -> frozendict[RefRec, tuple[RefDetector, ...]]:
        rec_to_detectors: dict[RefRec, list[RefDetector]] = {
            rec: [] for rec in self.recs
        }
        for detector in self.detectors:
            for ref_target in detector.targets:
                assert isinstance(ref_target, RefRec)
                rec_to_detectors[ref_target].append(detector)
        return frozendict(
            {
                ref_rec: tuple(detectors)
                for ref_rec, detectors in rec_to_detectors.items()
            }
        )

    @functools.cached_property
    def instruction_to_index(self) -> frozendict[RefInstruction, int]:
        return frozendict(
            {instruction: index for index, instruction in enumerate(self.instructions)}
        )

    @functools.cached_property
    def instruction_rec_biases(self) -> tuple[int, ...]:
        rec_bias: int = 0
        rec_biases: list[int] = []
        for instruction in self.instructions:
            rec_biases.append(rec_bias)
            rec_bias += instruction.num_measurements
        return tuple(rec_biases)

    @functools.cached_property
    def instruction_to_rec_bias(self) -> frozendict[RefInstruction, int]:
        return frozendict(
            {
                instruction: bias
                for instruction, bias in zip(
                    self.instructions, self.instruction_rec_biases
                )
            }
        )

    @functools.cached_property
    def detectors(self) -> tuple[RefDetector, ...]:
        detectors: list[RefDetector] = []
        for instruction in self.instructions:
            if instruction.name == "DETECTOR":
                detectors.append(instruction)
        return tuple(detectors)

    @functools.cached_property
    def detector_to_index(self) -> frozendict[RefDetector, int]:
        return frozendict(
            {detector: index for index, detector in enumerate(self.detectors)}
        )

    @functools.cached_property
    def stim_instructions(self) -> tuple[stim.CircuitInstruction, ...]:
        stim_instructions: list[stim.CircuitInstruction] = []
        for instruction in self.instructions:
            relative_targets = []
            for ref_target in instruction.targets:
                if isinstance(ref_target, RefRec):
                    relative_targets.append(
                        stim.target_rec(ref_target.rel_index(self, instruction))
                    )
                else:
                    relative_targets.append(ref_target)
            stim_instructions.append(
                stim.CircuitInstruction(
                    name=instruction.name,
                    targets=relative_targets,
                    gate_args=instruction.gate_args,
                    **extra_instruction_params(instruction),
                )
            )
        return tuple(stim_instructions)

    @functools.cached_property
    def ref_dem(self) -> "RefDetectorErrorModel":
        return RefDetectorErrorModel.of(self)

    def circuit(self) -> stim.Circuit:
        circuit = stim.Circuit()
        for stim_instruction in self.stim_instructions:
            circuit.append(stim_instruction)
        return circuit

    def __getitem__(self, index_or_slice: object) -> "RefInstruction | RefCircuit":
        if isinstance(index_or_slice, int):
            return self.instructions[index_or_slice]
        elif isinstance(index_or_slice, slice):
            return RefCircuit.of(self.instructions[index_or_slice])
        else:
            raise TypeError("Invalid key type")

    def __add__(self, other: "RefCircuit | RefInstruction") -> "RefCircuit":
        if isinstance(other, RefInstruction):
            return RefCircuit.of([*self.instructions, other])
        elif isinstance(other, RefCircuit):
            return RefCircuit.of([*self.instructions, *other.instructions])
        else:
            raise NotImplemented()

    def __radd__(self, other: "RefCircuit | RefInstruction") -> "RefCircuit":
        if isinstance(other, RefInstruction):
            return RefCircuit.of([other, *self.instructions])
        elif isinstance(other, RefCircuit):
            return RefCircuit.of([*other.instructions, *self.instructions])
        else:
            raise NotImplemented()

    def sanity_check(self) -> "RefCircuit":
        # check basic type
        assert isinstance(self.instructions, tuple)
        for instruction in self.instructions:
            assert isinstance(instruction, RefInstruction)
        # check if there are any duplicate instruction: id -> index
        existing_instruction: dict[RefInstruction, int] = {}
        for index, instruction in enumerate(self.instructions):
            assert instruction not in existing_instruction, (
                f"Duplicate instruction found instruction={instruction}, "
                + f"previous index={existing_instruction[instruction]}, "
                + f"current index={index}; if you want to repeat the same "
                + "circuit, considering using sub_circuit + sub_circuit.clone()"
            )
            existing_instruction[instruction] = index
        # check if all the references are valid in the current context
        rec_index_of: dict[RefRec, int] = {
            ref_rec: index for index, ref_rec in enumerate(self.recs)
        }
        assert len(rec_index_of) == len(self.recs), "has duplicate RefRec object"
        rec_hashes = set(hash(rec) for rec in self.recs)
        assert len(rec_hashes) == len(self.recs), "hash conflict"
        # check if the instructions only reference to the available RefRec objects
        # also check that all the RefRec objects refer to a valid Instruction
        previous_rec_set: set[RefRec] = set()
        previous_instruction_set: set[RefInstruction] = set()
        for instruction in self.instructions:
            for ref_target in instruction.targets:
                if isinstance(ref_target, RefRec):
                    assert ref_target in previous_rec_set, (
                        f"Reference to a RefRec object that does not appear previously, "
                        + f"ref_target={ref_target}"
                    )
                    assert ref_target.instruction in previous_instruction_set
            for ref_rec in instruction.recs:
                assert ref_rec.instruction == instruction  # must refer itself
                previous_rec_set.add(ref_rec)
            previous_instruction_set.add(instruction)
        # check that the number of measurements is correct
        for instruction, stim_instruction in zip(
            self.instructions, self.stim_instructions
        ):
            assert stim_instruction.num_measurements == instruction.num_measurements
        # check that all the detectors only targets RefRec objects
        for detector in self.detectors:
            for ref_target in detector.targets:
                assert isinstance(ref_target, RefRec)
        return self

    def clone(self) -> "RefCircuit":
        """
        make a deep copy of the RefCircuit object, creating branch new instructions
        with different ids, but keeping the circuit the same
        """
        new_instructions: list[RefInstruction] = []
        new_rec: list[RefRec] = []
        for instruction in self.instructions:
            new_instruction = RefInstruction(
                name=instruction.name,
                targets=tuple(
                    (
                        new_rec[self.rec_to_index[target]]
                        if isinstance(target, RefRec)
                        else target
                    )
                    for target in instruction.targets
                ),
                gate_args=instruction.gate_args,
                **extra_instruction_params(instruction),
            )
            for bias in range(instruction.num_measurements):
                reference_rec = RefRec(
                    instruction=new_instruction,
                    bias=bias,
                )
                new_instruction.recs.append(reference_rec)
                new_rec.append(reference_rec)
            new_instruction.recs.freeze()  # do not allow further edit
            new_instructions.append(new_instruction)
        return RefCircuit.of(new_instructions)

    def remove_noise_channels(
        self, keeping: Collection[RefInstruction]
    ) -> "RefCircuit":
        """
        Remove all the noise channels and their detectors from the circuit
        """
        new_instructions = list(self)
        deleting_indices: list[int] = []
        for index, instruction in enumerate(self.instructions):
            if not is_noise_channel_instruction(instruction):
                continue
            if instruction not in keeping:
                deleting_indices.append(index)
            # if the noise has measurement result, then they are heralded errors
            # remove the detectors associated with the heralded errors as well
            for ref_rec in instruction.recs:
                for detector in self.rec_to_detectors[ref_rec]:
                    assert (
                        len(detector.targets) == 1
                    ), "bug: detector of a heralded error has multiple targets"
                    if detector not in keeping:
                        deleting_indices.append(self.instruction_to_index[detector])
        assert len(set(deleting_indices)) == len(deleting_indices), "bug: duplicate"
        for index in sorted(deleting_indices, reverse=True):
            del new_instructions[index]
        return RefCircuit.of(new_instructions)


NOISE_CHANNEL_INSTRUCTION_WARNING_KEYWORDS: tuple[str, ...] = (
    "ERROR",
    "DEPOLARIZE",
    "HERALDED",
    "PAULI_CHANNEL",
)

# the available noise channels as of stim v1.15.0-dev
NOISE_CHANNEL_INSTRUCTION_NAMES: frozenset[str] = frozenset(
    {
        "CORRELATED_ERROR",
        "DEPOLARIZE1",
        "DEPOLARIZE2",
        "E",
        "ELSE_CORRELATED_ERROR",
        "HERALDED_ERASE",
        "HERALDED_PAULI_CHANNEL_1",
        "PAULI_CHANNEL_1",
        "PAULI_CHANNEL_2",
        "X_ERROR",
        "Y_ERROR",
        "Z_ERROR",
    }
)


def is_noise_channel_instruction(
    instruction: RefInstruction | stim.CircuitInstruction,
) -> bool:
    name: str = instruction.name
    if name in NOISE_CHANNEL_INSTRUCTION_NAMES:
        return True
    for keyword in NOISE_CHANNEL_INSTRUCTION_WARNING_KEYWORDS:
        if keyword in name:
            print(f"[warning] found keyword {keyword} in unknown instruction: {name}")
    return False


@dataclass(frozen=True)
class RefDemInstruction:
    type: str
    args: tuple[float, ...] = ()
    targets: tuple[int | stim.DemTarget | RefDetector, ...] = ()


@dataclass(frozen=True)
class RefDetectorErrorModel:
    instructions: tuple[RefDemInstruction, ...]
    ref_circuit: RefCircuit

    @staticmethod
    def of(
        ref_circuit: RefCircuit | None = None,
        dem: stim.DetectorErrorModel | None = None,
    ) -> "RefDetectorErrorModel":
        if dem is None:
            assert ref_circuit is not None, "circuit and dem cannot be both None"
            dem = ref_circuit.circuit().detector_error_model(
                approximate_disjoint_errors=True, flatten_loops=True
            )
        if ref_circuit is None:
            # create a mock ref_circuit just for the reference of detectors
            num_detectors = dem.num_detectors
            circuit = stim.Circuit()
            circuit.append("R", list(range(num_detectors)))
            circuit.append(
                "M", list(range(num_detectors))
            )  # some constant measurements
            for detector_id in range(num_detectors):
                circuit.append(
                    "DETECTOR", [stim.target_rec(detector_id - num_detectors)]
                )
            ref_circuit = RefCircuit.of(circuit)
        instructions: list[RefDemInstruction] = []
        for instruction in dem.flattened():
            ref_targets: list[int | stim.DemTarget | RefDetector] = []
            for target in instruction.targets_copy():
                if target.is_relative_detector_id():
                    ref_targets.append(ref_circuit.detectors[target.val])
                else:
                    ref_targets.append(target)
            # construct the instruction
            ref_instruction = RefDemInstruction(
                type=instruction.type,
                args=tuple(instruction.args_copy()),
                targets=tuple(ref_targets),
            )
            instructions.append(ref_instruction)
        return RefDetectorErrorModel(
            instructions=tuple(instructions), ref_circuit=ref_circuit
        )

    def dem(self, ref_circuit: RefCircuit | None = None) -> stim.DetectorErrorModel:
        if ref_circuit is None:
            ref_circuit = self.ref_circuit
        dem = stim.DetectorErrorModel()
        for instruction in self.instructions:
            dem.append(
                stim.DemInstruction(
                    type=instruction.type,
                    args=instruction.args,
                    targets=tuple(
                        (
                            stim.DemTarget.relative_detector_id(
                                ref_circuit.detector_to_index[target]
                            )
                            if isinstance(target, RefDetector)
                            else target
                        )
                        for target in instruction.targets
                    ),
                )
            )
        return dem

    @functools.cached_property
    def hyperedges(self) -> tuple["DemHyperedge", ...]:
        """
        we don't need to put all the hyperedges in the graph. If multiple hyperedges have
        the same detector set (incident vertices) but only differ by logical observable,
        then we only need to add the hyperedge with the highest probability and totally
        ignore others, as they will never be chosen in any MWPF solution anyway.
        """
        # detectors: dict[observables, probability]
        duplicate_edges: dict[frozenset[int], dict[frozenset[int], float]] = {}
        for instruction in self.instructions:
            if instruction.type == "error":
                assert (
                    len(instruction.args) == 1
                ), "error instruction must have 1 parameter of type float"
                probability = instruction.args[0]
                detectors: set[int] = set()
                observables: set[int] = set()
                for target in instruction.targets:
                    if isinstance(target, RefDetector):
                        detectors ^= {self.ref_circuit.detector_to_index[target]}
                    elif (
                        isinstance(target, stim.DemTarget)
                        and target.is_logical_observable_id()
                    ):
                        observables ^= {target.val}
                detectors = frozenset(detectors)
                observables = frozenset(observables)
                if detectors in duplicate_edges:
                    if observables in duplicate_edges[detectors]:
                        old_p = duplicate_edges[detectors][observables]
                        duplicate_edges[detectors][observables] = exclusive_probability(
                            old_p, probability
                        )
                    else:
                        duplicate_edges[detectors][observables] = probability
                else:
                    duplicate_edges[detectors] = {observables: probability}

        # detectors: (probability, observables)
        mapping: dict[frozenset[int], tuple[float, frozenset[int]]] = {}
        for detectors, values in duplicate_edges.items():
            # choosing the observables with the maximum probability
            observables, probability = max(values.items(), key=lambda x: x[1])
            mapping[detectors] = (probability, observables)
        return tuple(
            DemHyperedge(detectors, observables, probability)
            for detectors, (probability, observables) in mapping.items()
        )

    @functools.cached_property
    def hyperedges_detectors_set(self) -> frozenset[frozenset[int]]:
        return frozenset(dem_hyperedge.detectors for dem_hyperedge in self.hyperedges)

    @functools.cached_property
    def _dem(self) -> stim.DetectorErrorModel:
        return self.dem()

    @functools.cached_property
    def initializer(self) -> mwpf_incr.SolverInitializer:
        vertex_num = self._dem.num_detectors
        weighted_edges = [
            mwpf_incr.HyperEdge(
                dem_hyperedge.detectors,
                probability_to_weight(dem_hyperedge.probability),
            )
            for dem_hyperedge in self.hyperedges
        ]
        return mwpf_incr.SolverInitializer(vertex_num, weighted_edges)

    @functools.cached_property
    def predictor(self) -> "StaticPredictor":
        fault_masks = [
            sum(1 << k for k in dem_hyperedge.observables)
            for dem_hyperedge in self.hyperedges
        ]
        return StaticPredictor(
            fault_masks=fault_masks,
            num_dets=self._dem.num_detectors,
            num_obs=self._dem.num_observables,
        )


class Predictor(Protocol):
    """given the syndrome and the subgraph, predict the logical observable correction"""

    def syndrome_of(self, dets_bit_packed: np.ndarray) -> mwpf_incr.SyndromePattern: ...
    def prediction_of(
        self, syndrome: mwpf_incr.SyndromePattern, subgraph: Sequence[int]
    ) -> int: ...
    def num_detectors(self) -> int: ...
    def num_observables(self) -> int: ...

    def get_observable_bits(
        self, prediction: np.ndarray, shot_index: int = 0
    ) -> list[int]:
        return list(
            np.unpackbits(
                prediction[shot_index], count=self.num_observables(), bitorder="little"
            )
        )

    @property
    def is_dynamic(self) -> bool:
        return False


@dataclass(frozen=True)
class StaticPredictor(Predictor):
    fault_masks: list[int]
    num_dets: int
    num_obs: int

    def syndrome_of(self, dets_bit_packed: np.ndarray) -> mwpf_incr.SyndromePattern:
        defect_vertices: list[int] = list(
            np.flatnonzero(
                np.unpackbits(dets_bit_packed, count=self.num_dets, bitorder="little")
            )
        )
        return mwpf_incr.SyndromePattern(defect_vertices=defect_vertices)

    def prediction_of(
        self, _syndrome: mwpf_incr.SyndromePattern, subgraph: Sequence[int]
    ) -> int:
        return reduce(lambda x, y: x ^ y, (self.fault_masks[i] for i in subgraph), 0)

    def num_detectors(self) -> int:
        return self.num_dets

    def num_observables(self) -> int:
        return self.num_obs


@dataclass(frozen=True)
class DemHyperedge:
    detectors: frozenset[int]
    observables: frozenset[int]
    probability: float


WEIGHT_MAX: float = 1000  # larger than np.log(1e300)


def probability_to_weight(probability: float) -> float:
    if probability >= 1:
        return -WEIGHT_MAX
    return np.log((1 - probability) / probability)


def weight_to_probability(weight: float) -> float:
    return 1 / (1 + np.exp(weight))


def exclusive_probability(p1: float, p2: float) -> float:
    return p1 * (1 - p2) + p2 * (1 - p1)


# tag is introduced in stim 1.15
HAS_TAG = hasattr(stim.CircuitInstruction, "tag")


def extra_instruction_params(instruction: Any) -> dict[str, Any]:
    params: dict[str, Any] = {}
    if HAS_TAG:
        params["tag"] = instruction.tag
    return params



from common import *
import stim


def test_ref_circuit_simple():
    circuit_str = """\
R 0 1 2 3 4
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
MR 1 3
DETECTOR(1, 0) rec[-2]
DETECTOR(3, 0) rec[-1]
HERALDED_ERASE(0.01) 0
DETECTOR rec[-1]
M 0 2 4
DETECTOR(1, 1) rec[-2] rec[-3] rec[-6]
DETECTOR(3, 1) rec[-1] rec[-2] rec[-5]
OBSERVABLE_INCLUDE(0) rec[-1]\
"""
    circuit = stim.Circuit(circuit_str)
    print("######### original circuit #########")
    print(circuit)
    assert str(circuit) == circuit_str
    ref_circuit = mwpf.ref_circuit.RefCircuit.of(circuit)
    print("######### absolute indexed circuit #########")
    print(ref_circuit)
    assert (
        str(ref_circuit)
        == """\
R 0 1 2 3 4
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
MR 1 3
DETECTOR(1, 0) abs[0]
DETECTOR(3, 0) abs[1]
HERALDED_ERASE(0.01) 0
DETECTOR abs[2]
M 0 2 4
DETECTOR(1, 1) abs[4] abs[3] abs[0]
DETECTOR(3, 1) abs[5] abs[4] abs[1]
OBSERVABLE_INCLUDE(0) abs[5]\
"""
    )
    circuit_2 = ref_circuit.circuit()
    print("######### new circuit #########")
    print(circuit_2)
    assert isinstance(circuit_2, stim.Circuit)
    assert str(circuit_2) == circuit_str

    # now let's play with some of the functionalities
    # first, the measurement recording array
    recs = ref_circuit.recs
    assert len(recs) == 6
    assert [rec.instruction.index(ref_circuit) for rec in recs] == [6, 6, 9, 11, 11, 11]
    assert [rec.abs_index(ref_circuit) for rec in recs] == [0, 1, 2, 3, 4, 5]
    assert recs[0].rel_index(ref_circuit, ref_circuit[7]) == -2
    assert recs[1].rel_index(ref_circuit, ref_circuit[7]) == -1
    # second, the detectors
    detectors = ref_circuit.detectors
    assert len(detectors) == 5
    assert [detector.num_measurements for detector in detectors] == [0] * 5
    assert [len(detector.targets) for detector in detectors] == [1, 1, 1, 3, 3]
    assert [detector.index(ref_circuit) for detector in detectors] == [7, 8, 10, 12, 13]
    assert [ref_circuit.detector_to_index[detector] for detector in detectors] == list(
        range(5)
    )
    assert ref_circuit.rec_to_detectors[recs[0]] == (detectors[0], detectors[3])
    assert ref_circuit.rec_to_detectors[recs[1]] == (detectors[1], detectors[4])
    assert ref_circuit.rec_to_detectors[recs[2]] == (detectors[2],)
    assert ref_circuit.rec_to_detectors[recs[3]] == (detectors[3],)
    assert ref_circuit.rec_to_detectors[recs[4]] == (detectors[3], detectors[4])
    assert ref_circuit.rec_to_detectors[recs[5]] == (detectors[4],)
    # then the instructions
    assert [
        ref_circuit.instruction_to_index[instruction] for instruction in ref_circuit
    ] == list(range(15))
    # finally play with the dem
    original_dem = circuit.detector_error_model(
        approximate_disjoint_errors=True, flatten_loops=True
    ).flattened()
    print("######### original dem #########")
    print(original_dem)
    print("######### new dem #########")
    new_dem = ref_circuit.ref_dem.dem()
    print(new_dem)
    print(new_dem == original_dem)


def test_ref_circuit_heralded_erase_example():
    circuit_str_with_comment = """\
HERALDED_ERASE(0.01) 0
# Declare a flag detector based on the erasure
DETECTOR rec[-1]

# Erase qubit 2 with 2% probability
# Separately, erase qubit 3 with 2% probability
HERALDED_ERASE(0.02) 2 3

# Do an XXXX measurement
MPP X2*X3*X5*X7
# Apply partially-heralded noise to the two qubits
HERALDED_ERASE(0.01) 2 3 5 7
DEPOLARIZE1(0.0001) 2 3 5 7
# Repeat the XXXX measurement
MPP X2*X3*X5*X7
# Declare a detector comparing the two XXXX measurements
DETECTOR rec[-1] rec[-6]
# Declare flag detectors based on the erasures
DETECTOR rec[-2]
DETECTOR rec[-3]
DETECTOR rec[-4]
DETECTOR rec[-5]\
"""
    circuit_str = """\
HERALDED_ERASE(0.01) 0
DETECTOR rec[-1]
HERALDED_ERASE(0.02) 2 3
MPP X2*X3*X5*X7
HERALDED_ERASE(0.01) 2 3 5 7
DEPOLARIZE1(0.0001) 2 3 5 7
MPP X2*X3*X5*X7
DETECTOR rec[-1] rec[-6]
DETECTOR rec[-2]
DETECTOR rec[-3]
DETECTOR rec[-4]
DETECTOR rec[-5]\
"""
    circuit = stim.Circuit(circuit_str_with_comment)
    print("######### original circuit #########")
    print(circuit)
    assert str(circuit) == circuit_str
    ref_circuit = mwpf.ref_circuit.RefCircuit.of(circuit)
    print("######### absolute indexed circuit #########")
    print(ref_circuit)
    assert (
        str(ref_circuit)
        == """\
HERALDED_ERASE(0.01) 0
DETECTOR abs[0]
HERALDED_ERASE(0.02) 2 3
MPP X2*X3*X5*X7
HERALDED_ERASE(0.01) 2 3 5 7
DEPOLARIZE1(0.0001) 2 3 5 7
MPP X2*X3*X5*X7
DETECTOR abs[8] abs[3]
DETECTOR abs[7]
DETECTOR abs[6]
DETECTOR abs[5]
DETECTOR abs[4]\
"""
    )
    circuit_2 = ref_circuit.circuit()
    print("######### new circuit #########")
    print(circuit_2)
    assert isinstance(circuit_2, stim.Circuit)
    assert str(circuit_2) == circuit_str


def test_ref_circuit_heralded_pauli_channel_1_example():
    circuit_str_with_comment = """\
# With 10% probability perform a phase flip of qubit 0.
HERALDED_PAULI_CHANNEL_1(0, 0, 0, 0.1) 0
DETECTOR rec[-1]  # Include the herald in detectors available to the decoder

# With 20% probability perform a heralded dephasing of qubit 0.
HERALDED_PAULI_CHANNEL_1(0.1, 0, 0, 0.1) 0
DETECTOR rec[-1]

# Subject a Bell Pair to heralded noise.
MXX 0 1
MZZ 0 1
HERALDED_PAULI_CHANNEL_1(0.01, 0.02, 0.03, 0.04) 0 1
MXX 0 1
MZZ 0 1
DETECTOR rec[-1] rec[-5]  # Did ZZ stabilizer change?
DETECTOR rec[-2] rec[-6]  # Did XX stabilizer change?
DETECTOR rec[-3]    # Did the herald on qubit 1 fire?
DETECTOR rec[-4]    # Did the herald on qubit 0 fire?\
"""
    circuit_str = """\
HERALDED_PAULI_CHANNEL_1(0, 0, 0, 0.1) 0
DETECTOR rec[-1]
HERALDED_PAULI_CHANNEL_1(0.1, 0, 0, 0.1) 0
DETECTOR rec[-1]
MXX 0 1
MZZ 0 1
HERALDED_PAULI_CHANNEL_1(0.01, 0.02, 0.03, 0.04) 0 1
MXX 0 1
MZZ 0 1
DETECTOR rec[-1] rec[-5]
DETECTOR rec[-2] rec[-6]
DETECTOR rec[-3]
DETECTOR rec[-4]\
"""
    circuit = stim.Circuit(circuit_str_with_comment)
    print("######### original circuit #########")
    print(circuit)
    assert str(circuit) == circuit_str
    ref_circuit = mwpf.ref_circuit.RefCircuit.of(circuit)
    print("######### absolute indexed circuit #########")
    print(ref_circuit)
    assert (
        str(ref_circuit)
        == """\
HERALDED_PAULI_CHANNEL_1(0, 0, 0, 0.1) 0
DETECTOR abs[0]
HERALDED_PAULI_CHANNEL_1(0.1, 0, 0, 0.1) 0
DETECTOR abs[1]
MXX 0 1
MZZ 0 1
HERALDED_PAULI_CHANNEL_1(0.01, 0.02, 0.03, 0.04) 0 1
MXX 0 1
MZZ 0 1
DETECTOR abs[7] abs[3]
DETECTOR abs[6] abs[2]
DETECTOR abs[5]
DETECTOR abs[4]\
"""
    )
    circuit_2 = ref_circuit.circuit()
    print("######### new circuit #########")
    print(circuit_2)
    assert isinstance(circuit_2, stim.Circuit)
    assert str(circuit_2) == circuit_str


def test_adding_circuit():
    circuit_1 = stim.Circuit("MXX 0 1")
    circuit_2 = stim.Circuit("MZZ 2 3")
    circuit = circuit_1 + circuit_2
    print(circuit)
    assert str(circuit) == "MXX 0 1\nMZZ 2 3"
    ref_circuit_1 = mwpf.ref_circuit.RefCircuit.of(circuit_1)
    ref_circuit_2 = mwpf.ref_circuit.RefCircuit.of(circuit_2)
    ref_circuit = ref_circuit_1 + ref_circuit_2
    print(ref_circuit)
    assert str(ref_circuit) == "MXX 0 1\nMZZ 2 3"

    # adding the same circuit should work but not the ref circuits
    circuit = circuit_1 + circuit_1
    print(circuit)
    assert str(circuit) == "MXX 0 1 0 1"
    try:
        ref_circuit = ref_circuit_1 + ref_circuit_1
    except BaseException:
        ...
    else:
        raise Exception("it should panic")
    ref_circuit_1_clone = ref_circuit_1.clone()
    ref_circuit = ref_circuit_1 + ref_circuit_1_clone



from common import *


def test_flatten_config():
    code = mwpf.CodeCapacityColorCode(d=5, p=0.01)
    config = {
        "timeout": 6.0,
        "cluster_node_limit": 3,
    }
    mwpf.Solver(code.get_initializer(), config)


def test_legacy_config():
    code = mwpf.CodeCapacityColorCode(d=5, p=0.01)
    config = {
        "primal": {
            "timeout": 6.0,
            "cluster_node_limit": 3,
        }
    }
    mwpf.Solver(code.get_initializer(), config)
